function [communities] = improveModel(graph, communities, globalOutsideDensity,specialX)
%improveModel determines a better model of hyperbolic communities
% The expected input are the graph and a description of the communities in
% form of a struct array which for each communitiy contains the members
% (list of node indices of the given graph) and the paramType.
if ~exist('specialX', 'var')
    specialX = [];
end

explainedEdges = sparse(length(graph),length(graph));
coverGraph = sparse(length(graph),length(graph));

disp('phase 1: compute initial models and analyse overlaps');
for i = 1:length(communities)
    fprintf('.');
    comm = communities(i);
    [~, order] = degree_ord(graph(comm.members,comm.members) & ...
                            ~explainedEdges(comm.members,comm.members));    
    comm.members = comm.members(order);
    [ comm.p1, comm.p2, comm.LL ] = recoverCommunity(...
            graph(comm.members,comm.members) & ~explainedEdges(comm.members,comm.members),...
            comm.paramType, specialX);
    N = length(comm.members);
    comm.model = makeCommunity(N, comm.p1, comm.p2, getCondition(comm.paramType));
    explainedEdges(comm.members,comm.members) = explainedEdges(comm.members,comm.members) ...
                                                    | (graph(comm.members,comm.members) & comm.model);
    coverGraph(comm.members,comm.members) = coverGraph(comm.members,comm.members) | comm.model;
    comm.overlapsWith = [];
    for j = 1:length(communities)
        if j ~= i
            intersection  = intersect(communities(j).members,comm.members);
            if ~isempty(intersection)
                comm.overlapsWith(end+1) = j;
            end
        end
    end
    
    communities(i).members = comm.members;
    communities(i).p1 = comm.p1;
    communities(i).p2 = comm.p2;
    communities(i).LL = comm.LL;
    communities(i).model = comm.model;
    communities(i).overlapsWith = comm.overlapsWith;
end
fprintf('\n');

disp('phase 2: collect updatable communities ordered by LL');
[~, updatableCommunities] = sort([communities(:).LL], 'descend');
% we'll go to the global view for the outside-community part. the LLs
% computed previously don't compare
[communities.LL] = deal(-Inf);

disp('phase 3: iteratively update communities');
while ~isempty(updatableCommunities)
    fprintf('\t updating ');
    fprintf('%i ', updatableCommunities);
    fprintf('\n');
    updateList = [];
    explainedEdges = sparse(length(graph),length(graph));
    for commIdx = updatableCommunities
        fprintf('.');
        comm = communities(commIdx);
        % release the edges overlapping with the current model temporaryly
        explainedEdges(comm.members,comm.members) = ...
            explainedEdges(comm.members,comm.members) & ~comm.model;
        % erase the current community from the cover graph but keep all others
        coverGraph(comm.members,comm.members) = 0;
        for i = comm.overlapsWith
            coverGraph(communities(i).members, communities(i).members) = ...
                coverGraph(communities(i).members, communities(i).members) | communities(i).model;
        end

        %compute a new model for what we not already covered in a
        %previous iteration of the innermost for-loop
        [~, order] = degree_ord(graph(comm.members,comm.members) & ...
            ~explainedEdges(comm.members,comm.members));
        comm.members = comm.members(order);
        if globalOutsideDensity
            [ comm.p1, comm.p2, comm.LL ] = recoverCommunityInGraph(...
                    graph(comm.members,comm.members) & ...
                        ~explainedEdges(comm.members,comm.members), ...
                    comm.paramType, graph, coverGraph, coverGraph(comm.members,comm.members),specialX);
        else
            [ comm.p1, comm.p2, comm.LL ] = recoverCommunity(...
                    graph(comm.members,comm.members) & ...
                        ~explainedEdges(comm.members,comm.members), ...
                    comm.paramType);
        end
        N = length(comm.members);
        comm.model = makeCommunity(N, comm.p1, comm.p2, getCondition(comm.paramType));
        
        %update model if it improved and schedule updates for affected communities
        if comm.LL > communities(commIdx).LL 
            communities(commIdx).members = comm.members;
            communities(commIdx).p1 = comm.p1;
            communities(commIdx).p2 = comm.p2;
            communities(commIdx).LL = comm.LL;
            communities(commIdx).model = comm.model;    
            [~, orderOfUpdatableComms] = sort([communities(comm.overlapsWith).LL], 'descend');
            updateList = comm.overlapsWith(orderOfUpdatableComms);
        end

        % keep track of what is covered by the model
        explainedEdges(communities(commIdx).members,communities(commIdx).members) = ...
            explainedEdges(communities(commIdx).members,communities(commIdx).members) ...
            | (graph(communities(commIdx).members,communities(commIdx).members) & communities(commIdx).model);
        coverGraph(communities(commIdx).members,communities(commIdx).members) = ...
            coverGraph(communities(commIdx).members,communities(commIdx).members) ...
            | communities(commIdx).model;
        %figure;imagesc(graph(comm.members,comm.members) & comm.model & explainedEdges(comm.members,comm.members));
    end
    fprintf('\n');
    
    updatableCommunities = updateList;
    %ensure we make a full round of updates in each iteration
    %if ~isempty(updateList)
    %    [~, updatableCommunities] = sort([communities(:).LL], 'descend');
    %end
end

[~, commOrder] = sort([communities(:).LL], 'descend');
communities = communities(commOrder);
